/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.jproactor;

import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
//import java.util.concurrent.locks.ReentrantLock;
//import java.util.concurrent.atomic.AtomicBoolean;
//import java.util.concurrent.atomic.AtomicInteger;
//import java.util.concurrent.atomic.AtomicReference;

import java.util.LinkedList;

import EDU.oswego.cs.dl.util.concurrent.ReentrantLock;
import EDU.oswego.cs.dl.util.concurrent.SynchronizedBoolean;


/**
 * The <code>AsynchChannel</code> is an intermidiate interface that serves as
 * mediator between <code>Demultiplexor</code> and <code>AsynchHandler</code>
 * <p>
 * All implemented protocols should be based on this interface.
 *
 * @author <a href="mailto:libman@terabit.com.au">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 * 
 * @see IOOperation
 */
final public class AsynchChannel 
{

    /* Channel state constants */ 
    public final static int OPENED    = 1;
    public final static int CONNECTED = 2;
    public final static int CLOSING   = 3;
    public final static int CLOSED    = 4;


    /* ---------------------------------------------------------------------- */
    //
    // instance variables
    
    /** Internal lock for synchronization */
    private ReentrantLock m_lock = new ReentrantLock ();

    /** multiplexor reference */
    private Demultiplexor m_demultiplexor = null; 
    
    
    /** m_asynchChannel's channel */
    private SelectableChannel m_channel = null;
    
    /** This channel's handler, protocol instance */ 
    private AsynchHandler m_handler = null;
    
    /** channelState : OPENED, CONNECTED, CLOSING, CLOSED */
    private int m_channelState = OPENED;
    
    /** dispatch m_state can be DISP_NONE, DISP_QUEUED, DISP_ACTIVE */
//  private AtomicBoolean  m_dispatchActive = 
//  new AtomicBoolean(false); 
    private SynchronizedBoolean m_dispatchActive = new SynchronizedBoolean(false); 
    
    /** ready events as reported by selector */
    private int m_readyEvents = 0;
    
    /** event mask that is currently active in selector */
    private int m_remainEvents = 0;
    
    /** new mask that should be activated in selector after dispatch operation */
    private int m_interestEvents = 0;
    
    
    /** queue for the requested read operations */
//  private LinkedList<IOOperation> m_readQue = 
//  new LinkedList<IOOperation>();
    private LinkedList m_readQue = new LinkedList() ;

    /** queue for the requeted write operations */
//    private LinkedList<IOOperation> m_writeQue = 
//    	new LinkedList<IOOperation>();
    private LinkedList m_writeQue = new LinkedList() ;
    
    public IOStatistic m_readInfo  = new IOStatistic ();
    public IOStatistic m_writeInfo = new IOStatistic ();


    /* ---------------------------------------------------------------------- */
    
    /**
     * Creates <code>AsynchChannel</code> instance given specified 
     * <code>Demultiplexor</code>, <code>AsynchHandler</code> and 
     * <code>SelectableChannel</code> instances.
     *
     * @param m <code>Demultiplexor</code> instance this m_asynchChannel works with
     * @param p <code>AsynchHandler</code> protocol instance
     * @param sc <code>SelectableChannel</code> instance this m_asynchChannel operates on
     */
    public AsynchChannel( Demultiplexor m, AsynchHandler p, SelectableChannel sc )
    {
        this.m_demultiplexor = m;
        this.m_handler    = p;
        this.m_channel     = sc;
        
        m_demultiplexor.declareInterest (this);
    }

    /**
     * Returns protocol instance for this m_asynchChannel
     * 
     * @return protocol m_asynchChannel
     */
    AsynchHandler getHandler ()
    {
    	return m_handler;
    }

    /**
     * Completion dispather for this m_asynchChannel.
     * This method is called from Demultiplexor thread pool when
     * m_asynchChannel has completed operations.
     * 
     * It extacts all completioned operations from the completion queue
     * assosiated with this {@link #m_asynchChannel} and performs  upcalls
     * to the associated ProtocolHandlers
     */
    final void dispatchEvents (int readyEvents, int remainEvents)
    {
//    	m_lock.lock();
        try { m_lock.acquire(); }
        catch (InterruptedException e) { e.printStackTrace(); }

        boolean activeFlg = m_dispatchActive.commit(false,true);
    	try
    	{
    		m_readyEvents |= readyEvents;
    		m_remainEvents = remainEvents;
    		m_interestEvents &= ~readyEvents;
    		
    	   	if (!activeFlg)
        		return;
        	
  	   		int num = 0;
  	   	    
            // execute each IO operation only once then update interested events
            // mask it's possible that during IOOperation::onComplete() new
            // operations will be added
//    		do
//    		{
    	   		num = executeIO (m_readQue, (SelectionKey.OP_ACCEPT | SelectionKey.OP_READ));
    	   		num += executeIO (m_writeQue,(SelectionKey.OP_CONNECT | SelectionKey.OP_WRITE));
//    		}
//    		while (num > 0);
    		
    		if (m_channelState == CLOSING)
    		{
    			m_channelState = CLOSED;
    			m_handler.channelClosed( this );
    		}
    		else if (m_interestEvents != 0 &&
                     m_interestEvents != m_remainEvents )
    		{
    			m_demultiplexor.declareInterest(this);
    		}
     
    	}
    	catch (Exception e)
    	{
    		e.printStackTrace();
    	}
    	finally
    	{
    		if (activeFlg)
    		{
    			m_dispatchActive.set(false);
    		}
//    		m_lock.unlock();
            m_lock.release();
    		
    	}
    }
    /**
     * 
     * 
     */
//    private int executeIO (LinkedList<IOOperation> que, int opMask) 
    private int executeIO (LinkedList que, int opMask) 
    {
        int  num = 0;
        
        while((m_readyEvents & opMask) != 0 ||
              m_channelState == CLOSING || m_channelState == CLOSED) // m_channelState != OPENED)
        {
            // take operation from the queue
//            IOOperation op = que.poll();
//            IOOperation op = (IOOperation) que.poll();
            IOOperation op = (IOOperation) ((que.isEmpty()) ? null : que.remove(0));
            
        	if (op == null)
        	{
        		m_interestEvents &= ~opMask;
        		break;
        	}
        	
            boolean nextPossible = true;
            boolean completed = true;
            
          	if (m_channelState == CLOSING || m_channelState == CLOSED) // m_channelState != OPENED
        	{
        		op.setCancelled(null);
        	}
          	else if (op.executeSafe() == IOOperation.STATE_FINISHED)
          	{
          		if (op.getError() == null)
          		{
          			nextPossible = op.nextPossible();
                }
          	}
          	else
            {
            	nextPossible = false;
            	completed = false;
            }
          	
          	if (completed)
          	{
//                m_lock.unlock();
                m_lock.release();
                try
                {
                    op.onComplete();
                }
                catch (Exception e)
                {
                	e.printStackTrace();
                	nextPossible = true;
                	close ();
                }
                finally
                {
                	++num;
//                	m_lock.lock();
                    try { m_lock.acquire(); }
                    catch (InterruptedException e) { e.printStackTrace(); }
                }
          	}
          	else
          	{	
          		 // we are here if IOOperation was not executed
                 m_interestEvents |= op.m_type;
            	 m_readyEvents &= ~opMask;
            	 que.addFirst(op);
            	 break;
            }
          	if (!nextPossible)
          		break;
        }
        
        return num;
    }
    
    
   
    
    public void close() 
    {
    	boolean flgInterest = false;
//    	m_lock.lock();
    	try { m_lock.acquire(); }
        catch (InterruptedException e) { e.printStackTrace(); }

        try
        {
            if (this.m_channelState == OPENED || this.m_channelState == CONNECTED)
            {
                this.m_channelState = CLOSING;
                flgInterest = true;
            }
        }
    	finally 
        {
//        	m_lock.unlock();
            m_lock.release();
            
            
        	if (flgInterest)   
        	{
        		m_demultiplexor.declareInterest(this);
        	}
        }
    }

    private void startOperation(IOOperation op, LinkedList que) throws Exception
    {
    	boolean flgInterest = false;
//        m_lock.lock();
        m_lock.acquire();
        
        try
        {
        	if (m_channelState == CLOSING || m_channelState == CLOSED) // (m_channelState != OPENED)
            {
                throw new Exception (
                    "AsynchChannel::startOperation: attempt to start IO on inactive channel");
            }
            
            op.m_state = IOOperation.STATE_QUEUED;
            que.addLast(op);
            
            int opType = op.m_type;

            switch (opType) 
        	{
			case IOOperation.OP_READ:
			case IOOperation.OP_ACCEPT:
				++m_readInfo.m_opStarted;
				break;
			case IOOperation.OP_WRITE:
			case IOOperation.OP_CONNECT:
				++m_writeInfo.m_opStarted;
				break;
			}

    
            m_interestEvents |= opType;
            
            if ((m_remainEvents & opType) != 0)
            	return;
            
            if (m_dispatchActive.get())
            	return;
            
            flgInterest = true;
        }
        finally 
        {
//        	m_lock.unlock();
            m_lock.release();
            
        	if (flgInterest)   
        	{
        		m_demultiplexor.declareInterest(this);
        	}
        }
        
    }
    
    /**
     * Request for read operation.
     * This is main interface to start asynchronous read operation.
     * 
     * @param buf buffer to read data to
     * @return OpRead operation - future result
	 * @throws exception if error occurs
     */ 
	public OpRead read( ByteBuffer buf ) throws Exception
	{
		OpRead op = new OpRead( this, buf, false );
 
		startOperation (op, m_readQue );
		
		return op;
	}
	
	/**
	 * Request to execute non-partitional read.
	 * Read exactly {@link ByteBuffer#limit()} bytes to the buffer.
	 * 
	 * @param buf 
     * @return OpRead operation - future result
	 * @throws exception if error occurs
	 */
	public OpRead read_n ( ByteBuffer buf ) throws Exception
	{
		OpRead op = new OpRead( this, buf, true );
 
		startOperation (op, m_readQue );
		
        return op;
	}

    /**
     * Requets for write operation.
     * This is main interface to start asynchronous write operation.
     * 
     * @param buf buffer to write data from
     * @return OpWrite operation - future result
     * @throws exception if error occurs
	 */ 
	public OpWrite write( ByteBuffer buf ) throws Exception
	{
		OpWrite op = new OpWrite( this, buf, false );

		startOperation (op, m_writeQue );
		
		return op;
	}
	
	/**
	 * Request to execute non-partitional write.
	 * Write exactly {@link ByteBuffer#limit()} bytes to the associated channel.
	 * 
	 * @param buf buffer to write data from
	 * @return OpWrite operation - future result
	 * @throws exception if error occurs
	 */
	public OpWrite write_n( ByteBuffer buf ) throws Exception
	{
		OpWrite op = new OpWrite( this, buf, true );

		startOperation (op, m_writeQue );
		
		return op;
	}

	/**
	 * Request to execute accept operation.
	 * 
	 * @return OpAccept operation - future result
	 * @throws Exception if error occurs
	 */
    public OpAccept accept() throws Exception
    {
        OpAccept op = new OpAccept( this );
        
        startOperation (op, m_readQue );
        
        return op;
    }

	/**
	 * Request to execute connect operation.
	 * 
	 * @param SocketAddress  address to connect
	 * @return OpConnect operation - future result
	 * @throws Exception if error occurs
	 */
	public OpConnect connect (SocketAddress addr) throws Exception
	{
		OpConnect op = new OpConnect (this, addr);
		    	
		startOperation (op, m_writeQue );
		
        return op;
	}
    
    /**
     * Schedules timeout action to be called back on the associated 
     * <code>AsynchHandler</code> instance.
     * <p>
     * Note: while IO operations will be delivered to the handler in a
     * serialized way, timer notifications might be delivered concurrently.    
     * 
     * @param delay delay in milliseconds before timeout action is to be executed. 
     * @return OpTimer operation - future result
     */
    public OpTimer startTimer( long delay )
    {
        if ( delay < 0 )
            throw new IllegalArgumentException( "Negative timeout delay" );

        OpTimer op = new OpTimer( m_handler, delay );
        m_demultiplexor.startTimer( op );

        return op;
    }

    public boolean isOpened()    { return m_channelState == OPENED ||
                                          m_channelState == CONNECTED; }
    public boolean isConnected() { return m_channelState == CONNECTED; }
    

    // -----------------------------------------------------------------------
    //
    // Package visible methods
   
    /** 
     * Returns channel associated with this m_asynchChannel.
     */
    SelectableChannel getChannel() 
    { 
    	return m_channel;
    }
    
    private void doClose ()
    {
        // this will automatically cancel all keys
//    	m_lock.lock();
        try { m_lock.acquire(); }
        catch (InterruptedException e) { e.printStackTrace(); }

        try
        {
        	if (m_channel != null)
        		m_channel.close();
        }
        catch (Exception xc)
        {
        	xc.printStackTrace();
        }
        finally
        {
        	m_channel = null;
        	m_remainEvents = IOOperation.OP_ALL;
        	m_interestEvents = 0;
//        	m_lock.unlock();
            m_lock.release();
        }
    }
  
    
    /**
     * 
     * @param  selector to declare/cancel m_asynchChannel interest
     * @return true if interesrt updated
     *         false - failure, m_asynchChannel should be closed
     */
    boolean updateInterest (Selector selector)
    {
     	boolean rc = false;
//      m_lock.lock();
     	try { m_lock.acquire(); }
        catch (InterruptedException e) { e.printStackTrace(); }

        try
        {
        	if (m_channelState == OPENED || m_channelState == CONNECTED)
        	{
    	       	m_remainEvents |= m_interestEvents;
	            m_interestEvents = 0;
	 			SelectionKey key = m_channel.keyFor (selector);
		
				if (key == null)
				{
					m_channel.configureBlocking( false );
					key = m_channel.register (selector, m_remainEvents, this);
				}
				else
				{
					key.interestOps (m_remainEvents);
				}
				rc = true;
			}
        }
        catch ( Exception e ) 
        { 
        	e.printStackTrace();
        	rc = false;
        }
        finally
        {
           	if (!rc)
        	{
        		doClose ();
        	}
//            m_lock.unlock ();
            m_lock.release ();
        }
        	
        return rc;
    }
    
    void setState( int channelState )
    {
        m_channelState = channelState;
    }

    /**
     * DEBUG method: display requested operations
     * @return
     */
    String displayRequestedOps()
    {
        StringBuffer sb = new StringBuffer();
    
//        m_lock.lock() ;
        try { m_lock.acquire() ; }
        catch (InterruptedException e) { e.printStackTrace(); }

        try
        {    
            sb.append(" Read=").append (m_readQue.size() );
            sb.append(" Write=").append(m_writeQue.size());
        }
        finally
        {
//          m_lock.unlock ();
            m_lock.release();
        }
        
        return sb.toString();
    }
}
