/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.jproactor;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;

/**
 * The <tt>read</tt> operation implementation.
 *
 * @author <a href="mailto:libman@terabit.comg">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
public class OpRead extends IOOperation
{
    /** data buffer */
    protected ByteBuffer m_buffer = null;

    /** 
     * The number of bytes processed by IO operation.
     * <p> <tt>-1</tt> if error occured or channel is closed
     */
    protected int m_bytesCompleted = 0;
    
    /**
     * This flag denotes that data needs to be read until buffer is full.
     * 
     * @see java.nio.ByteBuffer#remaning()
     */  
    private boolean m_flgExactly = false;

    /**
     * @return a number of bytes available after io operation was completed;
     *          <tt>-1</tt> if exception or end-of-channel happened
     */
    public int getBytesCompleted()
    {
        return m_bytesCompleted;
    }

    /**
     * Gets operations' data buffer.
     * 
     * @return data buffer
     */
    public ByteBuffer getBuffer()
    {
        return m_buffer;
    }
    
    /** 
     * Tells the difference between a number of requested and processed bytes,
     *  by IO operationm, at present time.
     * <p>Example 1: Requested to read <i>n</i> bytes, actually read was
     * <i>m</i>, so bytesRemaining is <i>n-m</i> 
     * <p>Example 2: Requested to write <i>n</i> bytes, actually written was
     * <i>m</i>, so bytesRemaining is <i>n-m</i> 
     *
     * @see java.nio.ByteBuffer#remaning()
     */
    public int remaining()
    {
        return m_buffer.remaining();
    }


    /* ---------------------------------------------------------------------- */
    
    /**
     * Creates read operation given specified m_asynchChannel, which has io channel,
     * data buffer where to read data.
     *
     * @param asynchChannel
     * @param buffer read buffer
     * @param exactly specifies whether operation must ensure that input buffer
     *         must be full, i.e. all available space in the buffer is filled in,
     *         before operation is considered to be completed
     */
    OpRead( AsynchChannel achannel, ByteBuffer rdBuf, boolean exactly ) throws Exception
    {
        super (IOOperation.OP_READ, achannel);

        if ( rdBuf == null )
            throw new NullPointerException( "ByteBuffer is null" );
        if ( rdBuf.remaining() == 0 )
            throw new Exception( "Requested 0 bytes to read" );

        this.m_buffer  = rdBuf;
        this.m_flgExactly = exactly;
        this.m_state = IOOperation.STATE_STARTED;
    }

    /**
     * Reads data from associated channel, when channel is ready for
     * non-blocking read.
     *
     * @see IOOperation#execute()
     */
    public void execute() throws Exception
    {
        //System.out.println( Thread.currentThread().getName() + ": Executing Read IO" );

        SocketChannel channel = (SocketChannel) m_asynchChannel.getChannel();
        
        int n = channel.read( m_buffer );
            
        if ( n < 0 )
        {                       
            m_state = IOOperation.STATE_FINISHED;
            return;
        }
		
        m_bytesCompleted += n;
        m_asynchChannel.m_readInfo.m_xferBytes += n;
		
        if ( n == 0)
        {
            return;
        }   
        
        if ( !m_flgExactly || this.remaining()== 0 )
        {
            m_state = IOOperation.STATE_FINISHED;
            return;
        }
        
        return;
    }

    /**
     * @return <tt>true</tt> if next write operation can be executed,
     *          <tt>false</tt> otherwise
     * 
     * @see IOOperations#nextPossible()
     */
    public boolean nextPossible ()
    {
        return this.remaining() == 0;
    }
    
    /**
     * Called when IO operation has completed - notifies protocol instance
     * that read operation has completed.
     * 
     * @see IOOperation#onComplete()
     */
    public void onComplete()
    {
        try
        {
            m_buffer.flip(); // prepare buffer for get operations
            m_asynchChannel.getHandler().readCompleted(this);
        }
        catch (Exception e)
        {
            System.out.println("OpRead::onComplete() caught an error: " + e.getMessage());
            e.printStackTrace();
        }
    }

}
