/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.testmpx;

import java.nio.ByteBuffer;
import java.nio.channels.SelectableChannel;
import java.text.MessageFormat;

import au.com.terabit.jproactor.*;


/**
 * The <code>EchoServerProtocol</code> class is a simple <tt>Echo</tt> server
 * that demostrates usage and abilities of <i>JavaProactor</i> pattern.
 *
 * @author <a href="libman@terabit.com.au">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
class EchoServerProtocol implements AsynchHandler
{

    /** Default buffer size - 16 KB */
    private final static int BUFFER_SIZE = 4096;  //*4;  // 1 << 14;
    
    ByteBuffer buffer1 = ByteBuffer.allocate( BUFFER_SIZE );
    //ByteBuffer buffer2  = ByteBuffer.allocate( BUFFER_SIZE );
    
    AsynchChannel         achannel = null; 
    EchoServerProtocolFactory factory=null;
    
    /**
     * Creates a server instance for the given socket channel and multiplexor.
     *    
     * @param m - <code>Multiplexor</code> instance to use as IO processor.
     * @param channel - protocol's socket channel
     */
    EchoServerProtocol( Demultiplexor m,  SelectableChannel channel, EchoServerProtocolFactory f)
    {
    	this.factory = f;
        this.achannel = new AsynchChannel( m, this, channel );
    }

    /**
     * Start this protocol state machine.
     * <p>
     * On accept of a new connectio a new protocol instance is created,
     * then <code>Acceptor</code> instance executes this method.
     * 
     * @see ProtocolAdaptor#start()
     */
    public void start()throws Exception
    {
        System.out.println( Thread.currentThread().getName() + ": EchoServer protocol started" );
        
        achannel.read( buffer1 );
        //handler.read( buffer2 );
        
    }

    public void channelAttached(AsynchChannel channel) throws Exception {}

    public void channelClosed(AsynchChannel channel) throws Exception
    {
    	IOStatistic rdInfo = achannel.m_readInfo;
    	IOStatistic wrInfo = achannel.m_writeInfo;
    	
        System.out.println(
                MessageFormat.format(
                    "Server Connection Closed : Reads={0}({1},{2},{3}) Writes={4}({5},{6},{7})\n",
                     new Object[] {
                         new Long( rdInfo.m_xferBytes ),
                         new Long( rdInfo.m_opStarted ),
                         new Long( rdInfo.m_opFinished ),
                         new Long( rdInfo.m_opCancelled ),
                         new Long( wrInfo.m_xferBytes ),
                         new Long( wrInfo.m_opStarted ),
                         new Long( wrInfo.m_opFinished ),
                         new Long( wrInfo.m_opCancelled )
                    }
                ) );

        factory.onProtocolFinished(rdInfo, wrInfo);
        
    }

    /**
     * Notifies that read operation was completed.
     * 
     * @param opRead - read operation
     * 
     * @see IOHandler#onReadFinished(OpRead)
     */
    public void readCompleted( OpRead opRead ) throws Exception
    {
        if ( opRead.getError() != null )
        {    
             //System.out.println( "EchoServer::readCompleted: " + 
             //                 opRead.getError().toString());
            
           
             achannel.close();
             return;
        }
        
        if ( opRead.getBytesCompleted () <= 0)
        {
             //System.out.println( "EchoServer::readCompleted: Peer closed " +
             //                   opRead.getBytesCompleted() );
             achannel.close();
             return;
        }
        
        ByteBuffer buffer = opRead.getBuffer();
        //buffer.flip();
        achannel.write(buffer);
    }

    /**
     * Notifies that write operation was completed.
     * 
     * @param opWrite - write operation
     * 
     * @see IOHandler#onWriteFinished(OpWrite)
     */
    public void writeCompleted(OpWrite opWrite) throws Exception
    {
        if ( opWrite.getError() != null )
        {    
            System.out.println( "EchoServer::writeCompleted: " + 
                              opWrite.getError().toString());
            achannel.close();
            return;
        }
        
        if ( opWrite.getBytesCompleted () <= 0)
        {
             System.out.println( "EchoServer::writeCompleted: Peer closed " +
                                opWrite.getBytesCompleted() );
             achannel.close();
             return;
        }
        
        ByteBuffer buffer = opWrite.getBuffer();
        buffer.clear();
        
        achannel.read(buffer);
    }
    
    /**
     * Notifies that accept operation has completed.
     * <p>
     * @throws Exception - this operation should not happen once server
     *          protocol has started.
     * 
     * @see IOHandler#acceptCompletedshed(OpAccept)
     */
    public void acceptCompleted( OpAccept opAccept ) throws Exception
    {
        throw new Exception ("EchoServer::acceptCompleted - illegal call");
    }

    /**
     * Notifies that connect operation has completed.
     * <p>
     * @throws Exception - this operation should not happen once server
     *          protocol has started.
     * 
     * @see IOHandler#connectCompletedshed(OpConnect)
     */
    public void connectCompleted( OpConnect op ) throws Exception
    {
        throw new Exception ("EchoServer::connectCompleted - illegal call");
    }

    public void timerExpired(OpTimer opTimer) throws Exception {}

}
